@@ -1,5 +1,7 @@
 use {
+    crate::tower_storage::{SavedTower, TowerStorage},
     solana_gossip::cluster_info::ClusterInfo,
+    solana_measure::measure::Measure,
     solana_poh::poh_recorder::PohRecorder,
     solana_runtime::bank_forks::BankForks,
     solana_sdk::{clock::Slot, transaction::Transaction},
@@ -13,6 +15,7 @@ pub enum VoteOp {
     PushVote {
         tx: Transaction,
         tower_slots: Vec<Slot>,
+        saved_tower: SavedTower,
     },
     RefreshVote {
         tx: Transaction,
@@ -23,11 +26,8 @@ pub enum VoteOp {
 impl VoteOp {
     fn tx(&self) -> &Transaction {
         match self {
-            VoteOp::PushVote { tx, tower_slots: _ } => tx,
-            VoteOp::RefreshVote {
-                tx,
-                last_voted_slot: _,
-            } => tx,
+            VoteOp::PushVote { tx, .. } => tx,
+            VoteOp::RefreshVote { tx, .. } => tx,
         }
     }
 }
@@ -41,6 +41,7 @@ impl VotingService {
         vote_receiver: Receiver<VoteOp>,
         cluster_info: Arc<ClusterInfo>,
         poh_recorder: Arc<Mutex<PohRecorder>>,
+        tower_storage: Arc<dyn TowerStorage>,
         bank_forks: Arc<RwLock<BankForks>>,
     ) -> Self {
         let thread_hdl = Builder::new()
@@ -49,7 +50,13 @@ impl VotingService {
                 for vote_op in vote_receiver.iter() {
                     let rooted_bank = bank_forks.read().unwrap().root_bank().clone();
                     let send_to_tpu_vote_port = rooted_bank.send_to_tpu_vote_port_enabled();
-                    Self::handle_vote(&cluster_info, &poh_recorder, vote_op, send_to_tpu_vote_port);
+                    Self::handle_vote(
+                        &cluster_info,
+                        &poh_recorder,
+                        tower_storage.as_ref(),
+                        vote_op,
+                        send_to_tpu_vote_port,
+                    );
                 }
             })
             .unwrap();
@@ -59,18 +66,31 @@ impl VotingService {
     pub fn handle_vote(
         cluster_info: &ClusterInfo,
         poh_recorder: &Mutex<PohRecorder>,
+        tower_storage: &dyn TowerStorage,
         vote_op: VoteOp,
         send_to_tpu_vote_port: bool,
     ) {
+        if let VoteOp::PushVote { saved_tower, .. } = &vote_op {
+            let mut measure = Measure::start("tower_save-ms");
+            if let Err(err) = tower_storage.store(saved_tower) {
+                error!("Unable to save tower to storage: {:?}", err);
+                std::process::exit(1);
+            }
+            measure.stop();
+            inc_new_counter_info!("tower_save-ms", measure.as_ms() as usize);
+        }
+
         let target_address = if send_to_tpu_vote_port {
             crate::banking_stage::next_leader_tpu_vote(cluster_info, poh_recorder)
         } else {
             crate::banking_stage::next_leader_tpu(cluster_info, poh_recorder)
         };
-        let _ = cluster_info.send_vote(vote_op.tx(), target_address);
+        let _ = cluster_info.send_transaction(vote_op.tx(), target_address);
 
         match vote_op {
-            VoteOp::PushVote { tx, tower_slots } => {
+            VoteOp::PushVote {
+                tx, tower_slots, ..
+            } => {
                 cluster_info.push_vote(&tower_slots, tx);
             }
             VoteOp::RefreshVote {
